'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var functions = require('./functions-b6202a08.js');
require('folktale/concurrency/task');
var R = require('ramda');
require('folktale/result');
require('ramda-maybe');
require('util');
var throwingFunctions = require('./throwingFunctions.js');
require('folktale/maybe');
var monadHelpers = require('./monadHelpers-e0c6950d.js');
var NamedTupleMap = _interopDefault(require('namedtuplemap'));

/**
 * Created by Andy Likuski on 2018.12.28
 * Copyright (c) 2018 Andy Likuski
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/**
 *
 * Code modified from memoize-immutable. Memoizes a function using a flatArg version of the args
 * @param {Function} fn Function expecting any number of arguments
 * @returns {function([*], Object)} A function expecting
 * an array of the normal args for the second argument
 * an object of flattened args for the first argument
 */

var memoize = function memoize(fn) {
  var cache = new NamedTupleMap();
  var memoized = R.curry(function (normalArgs, flatArg) {
    if (!cache.has(flatArg)) {
      var result = R.apply(fn, normalArgs);
      cache.set(flatArg, result);
      return result;
    }

    return cache.get(flatArg);
  }); // Give a meaningful displayName to the memoized function

  if (fn.name) {
    memoized.displayName = fn.name + 'Memoized';
  }

  return memoized;
};
/** *
 * Memomizes a function to a single argument function so that we can always NamedTupleMap for the cache.
 * In order for this to work all objects have to be flattened into one big object. This Cache won't
 * accept inner objects that have changed. So the function coverts three args like
 * {a: {wombat: 1, emu: 2}}, {b: {caracal: 1, serval: 2}}, 'hamster' to
 * {arg1.a: {wombat: 1, emu: 2}, arg2.b: {caracal: 1, serval: 2}, arg3: 'hamster}.
 * You can provide any depth of objects as arguments, but it will have performance penalties
 * Consider memoizedWith to filter out unimportant data from arguments
 * @param {Function} func A function with any number and type of args
 * @returns {Function} A function that expects the same args as func
 */


var memoized = function memoized(func) {
  return memoizedWith(function () {
    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return args;
  }, func);
};
/** *
 * Memomizes a function to a single argument function so that we can always NamedTupleMap for the cache.
 * In order for this to work all objects have to be flattened into one big object. This Cache won't
 * accept inner objects that have changed. So the function coverts three args like
 * {a: {wombat: 1, emu: 2}}, {b: {caracal: 1, serval: 2}}, 'hamster' to
 * {arg1.a: {wombat: 1, emu: 2}, arg2.b: {caracal: 1, serval: 2}, arg3: 'hamster}.
 * You can provide any depth of objects as arguments, but it will have performance penalties.
 * To simplify the arguments and remove incomparable things like functions, use argumentFilter.
 * @param {Function} argumentFilter Expects the same number of arguments as func and returns an equal
 * number in an array. Each argument can be filtered to remove expensive objects or functions. Make
 * sure not to filter out anything that is used by func, since it will not figure into the argument uniqueness
 * @param {Function} func A function with any number and type of args. argumentFilter must match it
 * @returns {Function} A function that expects the same args as func. This function is curried, expecting
 * whatever number of arguments func is declared with, so you can call it partially
 */

var memoizedWith = function memoizedWith(argumentFilter, func) {
  // memoize(func) returns the memoized function expecting args and the flattened obj
  var memo = memoize(func); // Returns a function expecting the args for fun

  return R.curryN(func.length, function () {
    for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
      args[_key2] = arguments[_key2];
    }

    // Function that flattens the original args and called memoizedFunc with the single flattened arg
    return R.compose(memo(args), functions.flattenObj, // Filter out unneeded parts of the arguments, or does nothing if argumentFilter is ...args => args
    R.apply(argumentFilter))(args);
  });
};

var isObject = R.compose(R.equals('Object'), R.type);
var allAreObjectsOrLists = R.compose(R.either(R.all(isObject), R.all(Array.isArray)), R.values);

var hasVersion = function hasVersion(versionLabel) {
  return function (obj) {
    return R.has(versionLabel, obj);
  };
};

var hasBoth = function hasBoth(versionLabels) {
  return function (obj) {
    return R.all(function (versionLabel) {
      return hasVersion(versionLabel)(obj);
    }, versionLabels);
  };
};

var isEqual = function isEqual(versionLabels) {
  return function (obj) {
    return R.both(hasBoth(versionLabels), R.compose(R.apply(R.equals), R.values))(obj);
  };
};

var markAdded = function markAdded(obj) {
  return R.compose(R.append(undefined), R.values)(obj);
}; // eslint-disable-line no-undefined


var markRemoved = function markRemoved(obj) {
  return R.compose(R.prepend(undefined), R.values)(obj);
}; // eslint-disable-line no-undefined


var isAddition = function isAddition(versionLabels) {
  return function (obj) {
    return R.both(hasVersion(R.head(versionLabels)), R.complement(hasVersion(R.last(versionLabels))))(obj);
  };
};

var isRemoval = function isRemoval(versionLabels) {
  return function (obj) {
    return R.both(R.complement(hasVersion(R.head(versionLabels))), hasVersion(R.last(versionLabels)))(obj);
  };
};
/**
 * Diffs objects and replaces differences with an object {__left: left value, __right: right value}
 * @param {[String]} Two values indicating the names of the versions. If left null then __left and __right
 * are used for the diff
 * @param {Object} l The left object to diff
 * @param {Object} r The right object to diff
 * @returns {Object} A deep diff of l and r. Equal values are removed, different values are replaced by an
 * object keyed by the two version strings with the corresponding values
 */


var objectDiff = R.curry(function (versionLabels, l, r) {
  var labels = R.defaultTo(['__left', '__right'], versionLabels);

  var _labels = functions._slicedToArray(labels, 2),
      left = _labels[0],
      right = _labels[1];

  return R.compose(function (result) {
    return R.map(R.cond([[function (obj) {
      return isAddition(labels)(obj);
    }, function (obj) {
      return markAdded(obj);
    }], [function (obj) {
      return isRemoval(labels)(obj);
    }, function (obj) {
      return markRemoved(obj);
    }], [function (obj) {
      return hasBoth(labels)(obj);
    }, function (obj) {
      return R.when(allAreObjectsOrLists, // If all are objects or list recurse
      // Lists necessary become objects keyed by index since some indices are removed
      function (o) {
        return R.compose( // Remove the left, right labels and recurse
        function (values) {
          return R.apply(objectDiff(versionLabels), values);
        }, R.values)(o);
      } // Otherwise we keep the left and right labels and we are done
      )(obj);
    }]]), result);
  }, function (result) {
    return R.reject(function (obj) {
      return isEqual(labels)(obj);
    }, result);
  }, R.useWith(R.mergeWith(R.merge), [R.map(R.objOf(left)), R.map(R.objOf(right))]))(l, r);
});
/**
 * Pretty-prints objectDiff
 * @param {[String]} versionLabels The two labels or null
 * @param {Object} l The left-side value
 * @param {Object} r The right-side value
 * @returns {String} The diff string
 */

var prettyPrintObjectDiff = function prettyPrintObjectDiff(versionLabels, l, r) {
  return R.compose(function (result) {
    return JSON.stringify(result, null, 2);
  }, objectDiff)(versionLabels, l, r);
};

exports.alwaysFunc = functions.alwaysFunc;
exports.applyDeep = functions.applyDeep;
exports.applyDeepAndMapObjs = functions.applyDeepAndMapObjs;
exports.camelCase = functions.camelCase;
exports.capitalize = functions.capitalize;
exports.chainObjToValues = functions.chainObjToValues;
exports.compact = functions.compact;
exports.compactEmpty = functions.compactEmpty;
exports.compactJoin = functions.compactJoin;
exports.duplicateKey = functions.duplicateKey;
exports.emptyToNull = functions.emptyToNull;
exports.eqStrPath = functions.eqStrPath;
exports.eqStrPathsAll = functions.eqStrPathsAll;
exports.filterObjToValues = functions.filterObjToValues;
exports.filterWithKeys = functions.filterWithKeys;
exports.findByParams = functions.findByParams;
exports.findMapped = functions.findMapped;
exports.findOne = functions.findOne;
exports.findOneValueByParams = functions.findOneValueByParams;
exports.flattenObj = functions.flattenObj;
exports.flattenObjUntil = functions.flattenObjUntil;
exports.fromPairsDeep = functions.fromPairsDeep;
exports.hasStrPath = functions.hasStrPath;
exports.idOrIdFromObj = functions.idOrIdFromObj;
exports.isObject = functions.isObject;
exports.keyStringToLensPath = functions.keyStringToLensPath;
exports.lowercase = functions.lowercase;
exports.mapDefault = functions.mapDefault;
exports.mapDefaultAndPrefixOthers = functions.mapDefaultAndPrefixOthers;
exports.mapKeys = functions.mapKeys;
exports.mapKeysAndValues = functions.mapKeysAndValues;
exports.mapKeysForLens = functions.mapKeysForLens;
exports.mapObjToValues = functions.mapObjToValues;
exports.mapProp = functions.mapProp;
exports.mapPropValueAsIndex = functions.mapPropValueAsIndex;
exports.mapToObjValue = functions.mapToObjValue;
exports.mergeAllWithKey = functions.mergeAllWithKey;
exports.mergeDeep = functions.mergeDeep;
exports.mergeDeepAll = functions.mergeDeepAll;
exports.mergeDeepWith = functions.mergeDeepWith;
exports.mergeDeepWithConcatArrays = functions.mergeDeepWithConcatArrays;
exports.mergeDeepWithRecurseArrayItems = functions.mergeDeepWithRecurseArrayItems;
exports.mergeDeepWithRecurseArrayItemsAndMapObjs = functions.mergeDeepWithRecurseArrayItemsAndMapObjs;
exports.mergeDeepWithRecurseArrayItemsByAndMergeObjectByRight = functions.mergeDeepWithRecurseArrayItemsByAndMergeObjectByRight;
exports.mergeDeepWithRecurseArrayItemsByRight = functions.mergeDeepWithRecurseArrayItemsByRight;
exports.moveToKeys = functions.moveToKeys;
exports.omitDeep = functions.omitDeep;
exports.omitDeepBy = functions.omitDeepBy;
exports.omitDeepPaths = functions.omitDeepPaths;
exports.onlyOne = functions.onlyOne;
exports.onlyOneValue = functions.onlyOneValue;
exports.orEmpty = functions.orEmpty;
exports.overDeep = functions.overDeep;
exports.pickDeepPaths = functions.pickDeepPaths;
exports.removeDuplicateObjectsByProp = functions.removeDuplicateObjectsByProp;
exports.renameKey = functions.renameKey;
exports.replaceValuesAtDepth = functions.replaceValuesAtDepth;
exports.replaceValuesAtDepthAndStringify = functions.replaceValuesAtDepthAndStringify;
exports.replaceValuesWithCountAtDepth = functions.replaceValuesWithCountAtDepth;
exports.replaceValuesWithCountAtDepthAndStringify = functions.replaceValuesWithCountAtDepthAndStringify;
exports.reqPath = functions.reqPath;
exports.reqPathPropEq = functions.reqPathPropEq;
exports.reqStrPath = functions.reqStrPath;
exports.splitAtInclusive = functions.splitAtInclusive;
exports.strPath = functions.strPath;
exports.strPathOr = functions.strPathOr;
exports.strPathOrNullOk = functions.strPathOrNullOk;
exports.toArrayIfNot = functions.toArrayIfNot;
exports.transformKeys = functions.transformKeys;
exports.unflattenObj = functions.unflattenObj;
exports.unflattenObjNoArrays = functions.unflattenObjNoArrays;
exports.findOneThrowing = throwingFunctions.findOneThrowing;
exports.findOneValueByParamsThrowing = throwingFunctions.findOneValueByParamsThrowing;
exports.mappedThrowIfResultError = throwingFunctions.mappedThrowIfResultError;
exports.onlyOneThrowing = throwingFunctions.onlyOneThrowing;
exports.onlyOneValueThrowing = throwingFunctions.onlyOneValueThrowing;
exports.reqPathPropEqThrowing = throwingFunctions.reqPathPropEqThrowing;
exports.reqPathThrowing = throwingFunctions.reqPathThrowing;
exports.reqStrPathThrowing = throwingFunctions.reqStrPathThrowing;
exports.throwIfResultError = throwingFunctions.throwIfResultError;
exports.throwIfSingleResultError = throwingFunctions.throwIfSingleResultError;
exports.chainExceptMapDeepestMDeep = monadHelpers.chainExceptMapDeepestMDeep;
exports.chainMDeep = monadHelpers.chainMDeep;
exports.composeWithChain = monadHelpers.composeWithChain;
exports.composeWithChainMDeep = monadHelpers.composeWithChainMDeep;
exports.composeWithMap = monadHelpers.composeWithMap;
exports.composeWithMapExceptChainDeepestMDeep = monadHelpers.composeWithMapExceptChainDeepestMDeep;
exports.composeWithMapMDeep = monadHelpers.composeWithMapMDeep;
exports.defaultRunConfig = monadHelpers.defaultRunConfig;
exports.defaultRunToResultConfig = monadHelpers.defaultRunToResultConfig;
exports.doMDeep = monadHelpers.doMDeep;
exports.doMDeepExceptDeepest = monadHelpers.doMDeepExceptDeepest;
exports.lift1stOf2ForMDeepMonad = monadHelpers.lift1stOf2ForMDeepMonad;
exports.mapExceptChainDeepestMDeep = monadHelpers.mapExceptChainDeepestMDeep;
exports.mapMDeep = monadHelpers.mapMDeep;
exports.mapMonadByConfig = monadHelpers.mapMonadByConfig;
exports.mapOrObjToNamedResponseAndInputs = monadHelpers.mapOrObjToNamedResponseAndInputs;
exports.mapResultMonadWithOtherInputs = monadHelpers.mapResultMonadWithOtherInputs;
exports.mapResultTaskWithOtherInputs = monadHelpers.mapResultTaskWithOtherInputs;
exports.mapToMergedResponseAndInputs = monadHelpers.mapToMergedResponseAndInputs;
exports.mapToMergedResponseAndInputsMDeep = monadHelpers.mapToMergedResponseAndInputsMDeep;
exports.mapToNamedPathAndInputs = monadHelpers.mapToNamedPathAndInputs;
exports.mapToNamedResponseAndInputs = monadHelpers.mapToNamedResponseAndInputs;
exports.mapToNamedResponseAndInputsMDeep = monadHelpers.mapToNamedResponseAndInputsMDeep;
exports.mapToPath = monadHelpers.mapToPath;
exports.mapToResponseAndInputs = monadHelpers.mapToResponseAndInputs;
exports.mapWithArgToPath = monadHelpers.mapWithArgToPath;
exports.objOfMLevelDeepListOfMonadsToListWithPairs = monadHelpers.objOfMLevelDeepListOfMonadsToListWithPairs;
exports.objOfMLevelDeepMonadsToListWithPairs = monadHelpers.objOfMLevelDeepMonadsToListWithPairs;
exports.pairsOfMLevelDeepListOfMonadsToListWithPairs = monadHelpers.pairsOfMLevelDeepListOfMonadsToListWithPairs;
exports.promiseToTask = monadHelpers.promiseToTask;
exports.resultTasksToResultObjTask = monadHelpers.resultTasksToResultObjTask;
exports.resultToTask = monadHelpers.resultToTask;
exports.resultToTaskNeedingResult = monadHelpers.resultToTaskNeedingResult;
exports.resultToTaskWithResult = monadHelpers.resultToTaskWithResult;
exports.resultsToResultObj = monadHelpers.resultsToResultObj;
exports.retryTask = monadHelpers.retryTask;
exports.sequenceBucketed = monadHelpers.sequenceBucketed;
exports.stringifyError = monadHelpers.stringifyError;
exports.taskToPromise = monadHelpers.taskToPromise;
exports.taskToResultTask = monadHelpers.taskToResultTask;
exports.toMergedResponseAndInputs = monadHelpers.toMergedResponseAndInputs;
exports.toNamedResponseAndInputs = monadHelpers.toNamedResponseAndInputs;
exports.traverseReduce = monadHelpers.traverseReduce;
exports.traverseReduceDeep = monadHelpers.traverseReduceDeep;
exports.traverseReduceDeepResults = monadHelpers.traverseReduceDeepResults;
exports.traverseReduceError = monadHelpers.traverseReduceError;
exports.traverseReduceResultError = monadHelpers.traverseReduceResultError;
exports.traverseReduceWhile = monadHelpers.traverseReduceWhile;
exports.traverseReduceWhileBucketed = monadHelpers.traverseReduceWhileBucketed;
exports.traverseReduceWhileBucketedTasks = monadHelpers.traverseReduceWhileBucketedTasks;
exports.waitAllBucketed = monadHelpers.waitAllBucketed;
exports.memoized = memoized;
exports.memoizedWith = memoizedWith;
exports.objectDiff = objectDiff;
exports.prettyPrintObjectDiff = prettyPrintObjectDiff;
//# sourceMappingURL=index.js.map
